"""Scope type definitions."""
from enum import Enum


class ScopeType(Enum):
    """Scope types."""

    CALIBRATION = "calibration"
    PRE_EXPOSURE = "pre_exposure"
    DURING_EXPOSURE = "during_exposure"
    POST_EXPOSURE = "post_exposure"



"""Assonant Sample data class."""
from typing import List, Optional

from ..data_handlers import Axis
from .non_instrument import NonInstrument


class Sample(NonInstrument):
    """Data class to handle all data required to define a sample."""

    name: str
    positions: Optional[List[Axis]] = []



"""Assonant data classes - Instruments submodule.

Data classes responsible for grouping data related to instrumental data
"""

from .bvs import BVS
from .detector import Detector
from .instrument import Instrument
from .mirror import Mirror
from .monochromator import Monochromator
from .slit import Slit

__all__ = [
    "Instrument",
    "Detector",
    "Mirror",
    "Monochromator",
    "Slit",
    "BVS",
]



"""Factory class to deal with NeXus Object Creation."""


from typing import Dict, List

from nexusformat.nexus import *

from assonant.data_classes import AssonantDataClass
from assonant.data_classes.data_handlers import (
    Axis,
    DataField,
    DataHandler,
    ExternalLink,
    TimeSeries,
)
from assonant.data_classes.instruments import BVS, Detector, Mirror, Monochromator, Slit
from assonant.data_classes.non_instruments import Entry, Sample

from .exceptions import NeXusObjectFactoryError


class NeXusObjectFactory:
    """NeXus Object Factory Class.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating NeXus object out of AssonantDataClass objects.
    """

    def __init__(self):
        self.classes_converted_to_group = [
            Entry,
            Sample,
            Detector,
            Monochromator,
            BVS,
            Mirror,
            Slit,
            TimeSeries,
        ]
        self.classes_converted_to_field = [ExternalLink, DataField, Axis]

    def create_nxobject(self, data_obj: AssonantDataClass, pack_into_nxroot: bool = False) -> NXobject:
        """Creates the respective Nexus object based on the passed AssonantDataClass specific type.

        Args:
            data_obj (AssonantDataClass): Data object which will be used for the
            NeXus object creation
            pack_into_nxroot (bool, optional): If returned object should be or not
            packed inside a NXroot. Defaults to False.

        Raises:
            NeXusObjectFactoryError: The 'obj' argument type does not fit any of the
            supported types

        Returns:
            NXobject: NeXus object respective to the data object passed as the
            data_obj argument
        """
        if isinstance(data_obj, Entry):
            print("Constructor Called: ExperimentEntry")
            nxobject = self._create_entry(data_obj)
        elif isinstance(data_obj, Sample):
            print("Constructor Called: Sample")
            nxobject = self._create_sample(data_obj)
        elif isinstance(data_obj, Detector):
            print("Constructor Called: Detector")
            nxobject = self._create_detector(data_obj)
        elif isinstance(data_obj, Monochromator):
            print("Constructor Called: Monochromator")
            nxobject = self._create_monochromator(data_obj)
        elif isinstance(data_obj, BVS):
            print("Constructor Called: BVS")
            nxobject = self._create_bvs(data_obj)
        elif isinstance(data_obj, Mirror):
            print("Constructor Called: Mirror")
            nxobject = self._create_mirror(data_obj)
        elif isinstance(data_obj, Slit):
            print("Constructor Called: Slit")
            nxobject = self._create_slit(data_obj)
        elif isinstance(data_obj, TimeSeries):
            print("Constructor Called: TimeSeries")
            nxobject = self._create_time_series(data_obj)
        elif isinstance(data_obj, DataField):
            print("Constructor Called: Datafield")
            nxobject = self._create_data_field(data_obj)
        elif isinstance(data_obj, ExternalLink):
            print("Constructor Called: ExternalLink")
            nxobject = self._create_external_link(data_obj)
        else:
            raise NeXusObjectFactoryError(
                f"NeXus object factory doesn't have an constructor method to deal with objects of type: {type(data_obj)}"
            )

        # Check if there are subgroups to be created based on special fields
        if hasattr(data_obj, "fields"):
            print("Constructor Called: 'fields' Dictionary")
            nxobject = self._create_collection_from_dict(nxobject, data_obj.fields)

        if hasattr(data_obj, "positions"):
            print("Constructor Called: Transformations")
            nxobject = self._create_transformations(nxobject, data_obj.positions)

        return NXroot(nxobject) if pack_into_nxroot is True else nxobject

    def _create_entry(self, entry: Entry) -> NXentry:
        """Private method to create the NeXus object respective to any Entry data class.

        Args:
            entry (Entry): Data object containing entry data

        Returns:
            NXentry: NeXus object respective to the Entry data object
        """
        nx_scope_name = "instrument"
        nxobject = NXentry(name=entry.scope_type.value)
        nxobject.insert(NXinstrument(name=nx_scope_name))

        for instrument in entry.instruments:
            nxobject[nx_scope_name].insert(self.create_nxobject(instrument))

        if entry.sample is not None:
            nxobject.insert(self.create_nxobject(entry.sample))

        return nxobject

    def _create_sample(self, sample: Sample) -> NXsample:
        """Private method to create the NeXus object respective to the Sample data class.

        Args:
            sample (Sample): Data object containing sample data

        Returns:
            NXsample: NeXus object respective to the Sample data object
        """
        nx_scope_name = "sample"
        nxobject = NXsample(name=nx_scope_name)
        nxobject["name"] = sample.name

        return nxobject

    def _create_detector(self, detector: Detector) -> NXdetector:
        """Private method to create the NeXus object respective to the Detector data class.

        Args:
            detector (Detector): Data object containing detector data

        Returns:
            NXdetector: NeXus object respective to the Detector data object
        """
        nxobject = NXdetector(name=detector.name)

        return nxobject

    def _create_monochromator(self, monochromator: Monochromator) -> NXmonochromator:
        """Private method to create the NeXus object respective to the Monochromator data class.

        Args:
            monochromator (Monochromator): Data object containing monochromator
            data

        Returns:
            NXmonochromator: NeXus object respective to the Monochromator data
            object
        """
        nxobject = NXmonochromator(name=monochromator.name)

        return nxobject

    def _create_bvs(self, bvs: BVS) -> NXdetector:
        """Private method to create the NeXus object respective to the BVS data class.

        Args:
            bvs (BVS): Data object containing BVS data

        Returns:
            NXdetector: NeXus object respective to the BVS data object
        """
        nxobject = NXdetector(name=bvs.name)

        return nxobject

    def _create_mirror(self, mirror: Mirror) -> NXmirror:
        """Private method to create the NeXus object respective to the Mirror data class.

        Args:
            mirror (Mirror): Data object containing mirror data

        Returns:
            NXmirror: NeXus object respective to the Mirror data object
        """
        nxobject = NXmirror(name=mirror.name)

        return nxobject

    def _create_slit(self, slit: Slit) -> NXslit:
        """Private method to create the NeXus object respective to the Slit data class.

        Args:
            slit (Slit): Data object containing slit data

        Returns:
            NXslit: NeXus object respective to the Slit data object
        """
        nxobject = NXslit(name=slit.name)

        return nxobject

    def _create_external_link(self, external_link: ExternalLink) -> NXlink:
        """Private method to create the NeXus object respective to the ExternalLink data class.

        Args:
            external_link (ExternalLink): Data object containing external link
            data

        Returns:
            NXlink: NeXus object respective to the ExternalLink data object
        """
        nxobject = NXlink(name=external_link.name, target=external_link.target_path, file=external_link.filepath)

        return nxobject

    def _create_time_series(self, time_series: TimeSeries) -> NXlog:
        """Private method to create the NeXus object respective to the TimeSeries data class.

        Args:
            time_series (TimeSeries): Data object containing time_series data

        Returns:
            NXlog: NeXus object respective to the TimeSeries data object
        """
        nxobject = NXlog(
            value=self.create_nxobject(time_series.value),
            time=self.create_nxobject(time_series.timestamps),
        )

        return nxobject

    def _create_data_field(self, data_field: DataField) -> NXfield:
        """Private method to create the NeXus object respective to the Field data class.

        Args:
            data_field (DataField): Data object containing a single field
            data with its respective unit

        Returns:
            NXfield: NeXus object respective to the Field data object
        """
        attrs = data_field.extra_metadata

        if data_field.unit is not None:
            attrs["unit"] = data_field.unit

        nxobject = NXfield(value=data_field.value, attrs=attrs)

        return nxobject

    def _create_transformations(self, nxobject: NXobject, positions: List[Axis]) -> NXobject:
        """Create a NXtransformations group to store positioning data passed as a list of Axis objects.

        Args:
            nxobject (NXobject): NeXus object containing current objects
            positions (List[Axis]): List of Axis objects containing each of them the positioning data related to a
            specific monitored axis.

        Returns:
           NXobject: Respective NXobject with a NXcollection group containing all fields passed.
        """
        if positions != []:
            nxtransformations = NXtransformations(name="transformations")
            for axis in positions:
                nxtransformations[axis.name] = self.create_nxobject(axis.value)
                if isinstance(nxtransformations[axis.name], NXfield):
                    nxtransformations[axis.name].attrs["transformation_type"] = axis.transformation_type.value
                elif isinstance(nxtransformations[axis.name], NXlog):
                    nxtransformations[axis.name]["value"].attrs["transformation_type"] = axis.transformation_type.value
                else:
                    raise NeXusObjectFactoryError(
                        f"Type {nxtransformations[axis.name]} is an invalid conversion type for 'value' field from Axis object"
                    )
            nxobject.insert(nxtransformations)
        return nxobject

    def _create_collection_from_dict(self, nxobject: NXobject, data_dict: Dict[str, DataHandler] = {}) -> NXobject:
        """Create a NXcollection group to store data passed on 'data' dictionary.

        Each key, value tuple from the dictionary will be converted in a field with name equal to the tuple key.

        Args:
            nxobject (NXobject): NeXus object containing current objects
            data_dict (Dict): Dictionary containing data to populate NXcollection group that will be created. Defaults
            to {}.

        Raises:
            NeXusObjectFactoryError: Happens when a not supported AssonantDataClass is used on "data" dict values.

        Returns:
            NXobject: Respective NXobject with a NXcollection group containing all fields passed.
        """
        if data_dict != {}:
            for key in data_dict.keys():
                if isinstance(data_dict[key], DataField):
                    nxobject[key] = self.create_nxobject(data_dict[key])
                elif isinstance(data_dict[key], TimeSeries):
                    nxobject[key] = self.create_nxobject(data_dict[key])
                else:
                    raise NeXusObjectFactoryError(f"Expected 'data_dict[{key}]' to be a DataField or a TimeSeries")
        return nxobject



"""Assonant class to handle writing data from AssonantDataClasses into files."""

import os
from typing import List

from assonant.data_classes import AssonantDataClass

from ._assonant_file_writer_interface import IAssonantFileWriter
from ._file_writer_factory import FileWriterFactory


class AssonantFileWriter(IAssonantFileWriter):
    """Assonant File Writer. Wrapper class to deal with writing data from AssonantDataClass into files.

    Wrapper class that abstracts all process related to creating specific file writers,
    writing data on files following the requirements of the given format.
    """

    _factory = FileWriterFactory()

    def __init__(self, file_format: str):
        # Persist file format
        self.file_format = file_format

        # Create specific file writer based on chosen file format
        self.file_writer = self._factory.create_file_writer(file_format)

    def write_data(self, filepath: str, filename: str, data: AssonantDataClass):
        """Method for writing data using the specific FileWriter respective to define file format.

        If fail to save on target filepath, in order to avoid losing acquired data, it will try
        to save locally in current working directory.

        Args:
            filepath (str): Path where file will be saved.
            filename (str): Name that will be given to file.
            data (AssonantDataClass): AssonantDataClass that contains data to be saved in file.
        """
        try:
            self.file_writer.write_data(filepath, filename, data)
        except Exception as e:
            workdir_filepath = os.path.join(os.getcwd())
            print(
                f"Failed to save data at {filepath} due the following exception: {repr(e)}. Trying to save file locally in current work directory..."
            )
            try:
                self.file_writer.write_data(workdir_filepath, filename, data)
                print(f"File saved at current work directory (path: {workdir_filepath})")
            except Exception as e:
                print("Failed to save data locally in current work directory. Data lost.")
                raise e

    def get_supported_file_formats(self) -> List[str]:
        """Getter method for current supported file formats.

        Returns:
            List[str]: List containing curretly supported file formats.
        """
        return self._factory.get_supported_file_formats()



"""Assonant Entry data class."""
from typing import List, Optional

from ..instruments import Instrument
from ..types import ScopeType
from .non_instrument import NonInstrument
from .sample import Sample


class Entry(NonInstrument):
    """Data classes that wraps data into a logical/temporal scope related to the experiment.

    Entries are used to group and represent data in a defined temporal/logical scope of the
    experiment, which is directly define by the field "scope_type". e.g: calibration, pre-exposition.
    """

    scope_type: ScopeType
    instruments: Optional[List[Instrument]] = []
    sample: Optional[Sample] = None



"""Assonant DataField class."""
from typing import Dict, List, Optional, Type, Union

import numpy as np

from .data_handler import DataHandler


class DataField(DataHandler):
    """Data class to handle any type of base data, such as integers, floats, numpy arrays and lists."""

    value: Union[int, float, str, List, Type[np.ndarray]]
    unit: Optional[str] = None
    extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray]]]] = {}



import os

from nexusformat.nexus import nxopen

from assonant._nexus import NeXusObjectFactory
from assonant.data_classes import AssonantDataClass

from ._assonant_file_writer_interface import IAssonantFileWriter
from .exceptions import AssonantFileWriterError


class NexusFileWriter(IAssonantFileWriter):
    """NeXus File Writer.

    File Writer specialized on accessing data from AssonantDataClasses and
    write into a HDF5 file following NeXus data model standards.
    """

    def __init__(self):
        self.nexus_factory = NeXusObjectFactory()
        self.file_extension = ".nxs"

    def write_data(self, filepath: str, filename: str, data: AssonantDataClass):
        """Method for writing data into a NeXus file.

        File will be created if it doesn't exist yet, otherwise data will be add to NXroot object of the NeXus file.

        Args:
            filepath (str): Path where NeXus file will be saved.
            filename (str): Name that will be given to file.
            data (AssonantDataClass): AssonantDataClass that contains data
            to be saved in file.
        """
        nexus_filepath = os.path.join(filepath, os.path.splitext(filename)[0]) + self.file_extension
        print(nexus_filepath)
        try:
            with nxopen(nexus_filepath, mode="a").nxfile as nxf:
                nxobject = self.nexus_factory.create_nxobject(data, pack_into_nxroot=True)
                nxf.update(nxobject)
        except Exception as e:
            raise AssonantFileWriterError(
                f"NeXusFileWriter failed to write data into {filepath} due to the following exception: {repr(e)}"
            )



"""Assonant Axis data handler."""
from typing import Union

from ..types import TransformationType
from .data_field import DataField
from .data_handler import DataHandler
from .time_series import TimeSeries


class Axis(DataHandler):
    """Data class to handle data related to an axis position.

    When axis position is static the 'value' field must contain a single value and when it varies over time, it
    should be an array or TimeSeries.
    """

    name: str
    transformation_type: TransformationType
    value: Union[DataField, TimeSeries]



"""Exceptions triggered to warn about internal errors in Assonant Data Sender module."""


class AssonantDataSenderError(Exception):
    """Assonant Data Sender Error.

    Custom Exception to warn when any error occurs inside AssonantDataSender
    """

    pass



"""Assonant Beam Visualiation System (BVS) data class."""
from .instrument import Instrument


class BVS(Instrument):
    """Data class to handle all data required to define a Beam Visualization System (BVS)."""



"""Assonant Instrument abstract class."""
from typing import Dict, List, Optional

from ..assonant_data_class import AssonantDataClass
from ..data_handlers import Axis, DataHandler


# TODO: Make this class abstract
class Instrument(AssonantDataClass):
    """Abstract class that creates the base common requirements to define an Assonant Instrument."""

    name: str
    positions: Optional[List[Axis]] = []
    fields: Optional[Dict[str, DataHandler]] = {}



"""Assonant Non-Instrument abstract class."""
from typing import Dict, Optional

from ..assonant_data_class import AssonantDataClass
from ..data_handlers import DataHandler


# TODO: Make this class abstract
class NonInstrument(AssonantDataClass):
    """Abstract class that creates the base common requirements to define an Assonant Non-Instrument."""

    fields: Optional[Dict[str, DataHandler]] = {}



"""Assonant data classes - Non-instruments submodule.

Data classes responsible for grouping data related to non-instrumental data
"""
from .entry import Entry
from .non_instrument import NonInstrument
from .sample import Sample

__all__ = [
    "Entry",
    "NonInstrument",
    "Sample",
]



"""Assonant data classes - Events submodule.

Event classes responsible for standardizing how specific signalizations are done
in Assonant.
"""

from .assonant_event import AssonantEvent
from .experiment_done import ExperimentDone

__all__ = ["AssonantEvent", "ExperimentDone"]
